#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import matplotlib.pyplot as plt
import numpy as np
import glob
import os
import matplotlib as mpl
from scipy.interpolate import make_interp_spline

params = {
    'text.latex.preamble': r'\usepackage{amsmath}',
    'font.size': 14, 'text.usetex': True,
    'axes.formatter.use_mathtext': True
}

plt.rcParams.update(params)
plt.rcParams["figure.figsize"] = (7.5, 11.25)
plt.rcParams['axes.grid'] = True
plt.rc('xtick', labelsize=14)
plt.rc('ytick', labelsize=14)

font = {'family': 'serif',
        'serif':  'cmr10'
        }

plt.rc('font', **font)
plt.rc('xtick', direction='in')
plt.rc('ytick', direction='in')
plt.rc('axes', unicode_minus=False)


def main():

    caseList = [
        'data/R01/',
        'data/R05/',
        'data/R12/'
    ]

    expDataFile = [
        'data/experimentalData/R01_Force.txt',
        'data/experimentalData/R05_Force.txt',
        'data/experimentalData/R12_Force.txt'
    ]

    imgprefix = 'images/'

    n_lines = 4
    cmap = mpl.colormaps['viridis']
    colors = cmap(np.linspace(0, 0.8, n_lines))
    tsize = 18
    lsize = 16

    lagVal = True  # Whether to time shift the numerical results
    perioda = [
        1.14,
        0.88,
        1.42
    ]
    p11Peak = [
        10.32,
        10.90,
        7.545
    ]
    perDiv = [
        1000,
        1500,
        2000
    ]
    dt = np.divide(perioda, perDiv)
    intStart = [
        9.18,
        10.02,
        6.13
    ]
    intEnd = [
        11.46,
        11.78,
        8.97
    ]

    for i in range(len(caseList)):

        ## Extract probe data
        dirWave = caseList[i]+'wave'
        nProbes, dataInfo, probeData = readProbesDirectory('waveProbes',dirWave)
        probeTime = probeData[:, 0]
        heightData = probeData[:, 3::3]
        
        ## Extract the force and moment data - account for symmetry and initial forces
        dirForce = caseList[i]+'sphere'
        dataInfo, forceData = readForcesDirectory('forcesSphere',dirForce)
        simTime = forceData[:,0]
        forcex = forceData[:,1]*2.0
        forcez = (forceData[:,3]-forceData[5,3])*2.0
        
        dataInfo, momentData = readMomentsDirectory('forcesSphere',dirForce)
        momenty = momentData[:,2]*2.0
     
        if any(probeTime != simTime):
            raise Exception('Times do not match!')

        ## Read in the experimental data and find interesting times
        expTime, expWave, expForcex, expForcez, expMomenty = getExampleForceMomentData(expDataFile[i])
        indExp = findTimes(expTime, intStart[i], intEnd[i])

        ## Determine the phase shift based on the peaks
        if lagVal == True:
            # Sync data based on probe 11
            timeRangeMinSync = p11Peak[i] - perioda[i]/4
            timeRangeMaxSync = p11Peak[i] + perioda[i]/4
            inds = findTimes(probeTime, timeRangeMinSync, timeRangeMaxSync)
            redTime = probeTime[inds]
            redData = heightData[inds, 11]
            maxInd = np.argmax(redData)
            lag = round((redTime[maxInd] - p11Peak[i])/dt[i])
        else:
            lag = 0

        ## Print out the lag value
        print('The lag for case ' +caseList[i][5:8]+' is: '+str('{:2.3f}'.format(-lag*dt[i])))

        ## Plot the full 12 seconds
        plt.figure(dpi=300)
        fig, ax = plt.subplots(4, 1, layout='constrained')
        ax[0].plot(expTime, expWave, color='red',linewidth=2, linestyle='dashed')
        ax[1].plot(expTime, expForcex, color='red',linewidth=2, linestyle='dashed')
        ax[2].plot(expTime, expForcez, color='red',linewidth=2, linestyle='dashed')
        ax[3].plot(expTime, expMomenty, color='red',label='Experimental', linewidth=2, linestyle='dashed')

        ## Shift the data based on the peaks
        probeTimeShifted, heightDataShifted, forcexShifted, forcezShifted, momentyShifted = syncDataSeriesLagForceMoment(probeTime,heightData[:,11],forcex,forcez,momenty,lag)

        ## Find interesting times
        ind = findTimes(probeTimeShifted, intStart[i], intEnd[i])

        ## Extract max absolute values
        magExpForcex = max(abs(max(expForcex[indExp])),abs(min(expForcex[indExp])))
        magExpForcez = max(abs(max(expForcez[indExp])),abs(min(expForcez[indExp])))
        magExpMomenty = max(abs(max(expMomenty[indExp])),abs(min(expMomenty[indExp])))
        
        magForcex = max(abs(max(forcexShifted[ind])),abs(min(forcexShifted[ind])))
        magForcez = max(abs(max(forcezShifted[ind])),abs(min(forcezShifted[ind])))
        magMomenty = max(abs(max(momentyShifted[ind])),abs(min(momentyShifted[ind])))
        
        ## Calculate some error metrics
        forcexError = np.linalg.norm(magForcex-magExpForcex)/np.linalg.norm(magExpForcex)*100
        forcezError = np.linalg.norm(magForcez-magExpForcez)/np.linalg.norm(magExpForcez)*100
        momentyError = np.linalg.norm(magMomenty-magExpMomenty)/np.linalg.norm(magExpMomenty)*100

        ## Fit data and downsample to match experimentl data to calculate trace error
        syncTime = expTime[indExp]
        splineFitForcex = bsplineInterp(probeTimeShifted, forcexShifted)
        forcexSpline = splineFitForcex(syncTime)
        forcexTraceError = np.linalg.norm(forcexSpline-expForcex[indExp])/np.linalg.norm(expForcex[indExp])*100
        
        splineFitForcez = bsplineInterp(probeTimeShifted, forcezShifted)
        forcezSpline = splineFitForcez(syncTime)
        forcezTraceError = np.linalg.norm(forcezSpline-expForcez[indExp])/np.linalg.norm(expForcez[indExp])*100
        
        splineFitMomenty = bsplineInterp(probeTimeShifted, momentyShifted)
        momentySpline = splineFitMomenty(syncTime)
        momentyTraceError = np.linalg.norm(momentySpline-expMomenty[indExp])/np.linalg.norm(expMomenty[indExp])*100
        
        ## Print out the maximum magnitudes
        print('The maximum magnitude of x-force for case '+caseList[i][5:8]+' is: '+str('{:2.6f}'.format(magForcex)))
        print('The maximum magnitude of z-force for case '+caseList[i][5:8]+' is: '+str('{:2.6f}'.format(magForcez)))
        print('The maximum magnitude of y-moment for case '+caseList[i][5:8]+' is: '+str('{:2.6f}'.format(magMomenty)))

        ## Print out some error metrics
        print('The x-force error for case '+caseList[i][5:8]+' is: '+str('{:2.6f}'.format(forcexError)))
        print('The z-force error for case '+caseList[i][5:8]+' is: '+str('{:2.6f}'.format(forcezError)))
        print('The y-moment error for case '+caseList[i][5:8]+' is: '+str('{:2.6f}'.format(momentyError)))
        
        print('The x-force trace error for case '+caseList[i][5:8]+' is: '+str('{:2.6f}'.format(forcexTraceError)))
        print('The z-force trace error for case '+caseList[i][5:8]+' is: '+str('{:2.6f}'.format(forcezTraceError)))
        print('The y-moment trace error for case '+caseList[i][5:8]+' is: '+str('{:2.6f}'.format(momentyTraceError)))
        
        ax[0].plot(probeTimeShifted, heightDataShifted,color=colors[0], linewidth=2)
        ax[1].plot(probeTimeShifted, forcexShifted,color=colors[0], linewidth=2)
        ax[2].plot(probeTimeShifted, forcezShifted,color=colors[0], linewidth=2)
        ax[3].plot(probeTimeShifted, momentyShifted,color=colors[0], label='Numerical', linewidth=2)
                    
        plt.setp(ax, xlim=(0, 12))
        plt.locator_params(axis='y', nbins=6)
        ax[0].set_title('$\eta$ at Probe 11',fontsize=tsize)
        ax[1].set_title('$x$-Force',fontsize=tsize)
        ax[2].set_title('$z$-Force',fontsize=tsize)
        ax[3].set_title('$y$-Moment',fontsize=tsize)
        ax[0].set_ylabel('[m]',fontsize=tsize)
        ax[1].set_ylabel('[N]',fontsize=tsize)
        ax[2].set_ylabel('[N]',fontsize=tsize)
        ax[3].set_ylabel('[N-m]',fontsize=tsize)
        fig.suptitle('Wave Surface Elevation, Forces, and Moment for $t \in [0,12]$',fontsize=tsize)
        ax[3].set_xlabel('Time [s]', fontsize=tsize)
        plt.legend(bbox_to_anchor=(0.5, -0.6), loc='upper center',borderaxespad=0, prop={'size': lsize})
        plt.tight_layout()
        plt.savefig(imgprefix+'Sphere_Force_Moment_' +caseList[i][5:8]+'.png', format='png')

        ## Plot the interesting time
        plt.figure(dpi=300)
        fig, ax = plt.subplots(4, 1, layout='constrained')
        ax[0].plot(expTime, expWave, color='red',linewidth=2, linestyle='dashed')
        ax[1].plot(expTime, expForcex, color='red',linewidth=2, linestyle='dashed')
        ax[2].plot(expTime, expForcez, color='red',linewidth=2, linestyle='dashed')
        ax[3].plot(expTime, expMomenty, color='red',label='Experimental', linewidth=2, linestyle='dashed')

        ## Shift the data based on the peaks
        probeTimeShifted, heightDataShifted, forcexShifted, forcezShifted, momentyShifted = syncDataSeriesLagForceMoment(probeTime,heightData[:,11],forcex,forcez,momenty,lag)

        ## Find interesting times
        ind = findTimes(probeTimeShifted, intStart[i], intEnd[i])
        
        ax[0].plot(probeTimeShifted, heightDataShifted,color=colors[0], linewidth=2)
        ax[1].plot(probeTimeShifted, forcexShifted,color=colors[0], linewidth=2)
        ax[2].plot(probeTimeShifted, forcezShifted,color=colors[0], linewidth=2)
        ax[3].plot(probeTimeShifted, momentyShifted,color=colors[0], label='Numerical', linewidth=2)

        plt.setp(ax, xlim=(intStart[i], intEnd[i]))
        ax[0].set_title('$\eta$ at Probe 11',fontsize=tsize)
        ax[1].set_title('$x$-Force',fontsize=tsize)
        ax[2].set_title('$z$-Force',fontsize=tsize)
        ax[3].set_title('$y$-Moment',fontsize=tsize)
        ax[0].set_ylabel('[m]',fontsize=tsize)
        ax[1].set_ylabel('[N]',fontsize=tsize)
        ax[2].set_ylabel('[N]',fontsize=tsize)
        ax[3].set_ylabel('[N-m]',fontsize=tsize)
        fig.suptitle('Wave Surface Elevation, Forces, and Moment for $t \in [$'+str(intStart[i])+','+str(intEnd[i])+'$]$',fontsize=tsize)
        ax[3].set_xlabel('Time [s]', fontsize=tsize)
        plt.legend(bbox_to_anchor=(0.5, -0.6), loc='upper center',borderaxespad=0, prop={'size': lsize})
        plt.tight_layout()
        plt.savefig(imgprefix+'Sphere_Force_Moment_' +caseList[i][5:8]+'_Interesting.png', format='png')


# Beginning of functions

def getFileList(datadir, prefix):
    datafiles = glob.glob(datadir+'/*/'+prefix+'.dat')

    nfiles = len(datafiles)

    if nfiles == 0:
        print('No files found for '+datadir+'/*/'+prefix+'.dat')

    ## Sort directories by time value (as opposed to alphabetical time 100 < 20)

    ## Get time valued subdirectories and make them floats for sorting
    datatimes = [dfile.replace(datadir+'/', '').replace('/'+prefix+'.dat', '') for dfile in datafiles]

    ## Check if filename"_time".dat exists, if so, use that one
    for i in range(len(datatimes)):
        dtime = datatimes[i]
        dfile = datafiles[i]

        testfile = dfile.replace('.dat', '_'+dtime+'.dat')
        if os.path.isfile(testfile):
            datafiles[i] = testfile

    ## Sort files on time
    if len(datafiles) > 1:
        datatimes = [float(t) for t in datatimes]
        ## Sort on time and throw away time array
        datafiles = [dfiles for dummy, dfiles in sorted(zip(datatimes, datafiles))]

    print(datafiles)

    return datafiles

def readData(nheader, ncolumns, datafiles):
    ## Empty data array
    data = np.empty((0, ncolumns))

    ## Read in data, appending each file
    for dfile in datafiles:
        with open(dfile) as fin:
            ## Skip header
            for i in range(nheader):
                fin.readline()

            ## Raw data
            dataTmp = fin.readlines()

            ## Formatted data
            ## Remove any parentheses and split
            data0 = np.array([[float(val) for val in np.array(line.replace('(', '').replace(')', '').split())] for line in dataTmp])

            ## Remove overlapping times
            ## WARNING HARD CODED to assume time is column 0
            if len(data) > 0:
                for iEntry in range(len(data[:, 0])):
                    if data[iEntry, 0] >= data0[0, 0]:
                        data = data[:iEntry, :]
                        break

            ## Append this onto data array
            data = np.vstack((data, data0))
    return data

def readForcesDirectory(dirName,directory):
    nheader=4  ## Number of header lines
    datadir=directory+'/postProcessing/'+dirName  ## Location of data to plot
    prefix ='force'
    ncolumns=10
    dataInfo=[ 'Time',
               'total_x', 'total_y', 'total_z',
               'pressure_x', 'pressure_y', 'pressure_z',
               'viscous_x', 'viscous_y', 'viscous_z',
             ]

    datafiles = getFileList(datadir,prefix)

    data = readData(nheader,ncolumns,datafiles)
   
    return dataInfo,data

def readMomentsDirectory(dirName,directory):
    nheader=4  ## Number of header lines
    datadir=directory+'/postProcessing/'+dirName  ## Location of data to plot
    prefix ='moment'
    ncolumns=10
    dataInfo=[ 'Time',
               'total_x', 'total_y', 'total_z',
               'pressure_x', 'pressure_y', 'pressure_z',
               'viscous_x', 'viscous_y', 'viscous_z',
             ]

    datafiles = getFileList(datadir,prefix)

    data = readData(nheader,ncolumns,datafiles)
   
    return dataInfo,data

def getProbeDataFull(expDataFile):
    headerLines = 2  ## This is hardcoded!
    probes = []
    with open(expDataFile) as fin:

        for i in range(headerLines):

            fin.readline()

        dataIn = fin.readlines()
        for line in dataIn:
            splitLn = [float(value) for value in line.split()]
            probes.append(splitLn)

    return np.asarray(probes)

def getProbeData(expDataFile):
    ## Extract the raw data
    probesFull = getProbeDataFull(expDataFile)

    ## Extract the time and mean values for each of the 12 probes
    probeTime = probesFull[:, 0]
    probeDataAvg = np.zeros((len(probesFull), 12))
    probeDataUpper = np.zeros((len(probesFull), 12))
    probeDataLower = np.zeros((len(probesFull), 12))
    for i in range(12):
        probeDataAvg[:, i] = probesFull[:, 1+3*(i)]
        probeDataUpper[:, i] = probesFull[:, 2+3*(i)]
        probeDataLower[:, i] = probesFull[:, 3+3*(i)]

    return probeTime, probeDataAvg, probeDataUpper, probeDataLower

def getExampleForceMomentData(expDataFile):
    headerLines = 1 ## This is hardcoded!
    data = []
    with open(expDataFile) as fin:
        
        for i in range(headerLines):
            
            fin.readline()
            
        dataIn = fin.readlines()
        for line in dataIn:
            splitLn = [float(value) for value in line.split()]
            data.append(splitLn)
        
    dataFull = np.asarray(data)
    dataTime = dataFull[:,0]
    dataEta = dataFull[:,1]
    dataFx = dataFull[:,2]
    dataFz = dataFull[:,4]
    dataMy = dataFull[:,6]
    
    return dataTime,dataEta,dataFx,dataFz,dataMy

def readProbesDirectory(dirName, directory):
    nheader = 0  ## Number of header lines
    datadir = directory+'/postProcessing/'+dirName  # Location of data to plot

    prefix = 'position'

    datafiles = getFileList(datadir, prefix)

    with open(datafiles[0]) as fin:
        for ii in range(nheader+1):
            dataline = fin.readline().replace('(', '').replace(')', '')
    ncolumns = len(dataline.split())

    # Position
    nprobes = int((ncolumns-1)/3)

    dataInfo = ['Time']
    for iprobe in range(nprobes):
        dataInfo.append('x_{}'.format(iprobe))
        dataInfo.append('y_{}'.format(iprobe))
        dataInfo.append('z_{}'.format(iprobe))

    data = readData(nheader, ncolumns, datafiles)

    return nprobes, dataInfo, data


def bsplineInterp(x, y, order=3, reltol=0.01):
    bspline = make_interp_spline(x, y, order)
    error = np.linalg.norm(bspline(x)-y)/np.linalg.norm(y)
    assert error <= reltol

    return bspline


def findTimes(time, timeRangeMin, timeRangeMax):
    ind = np.where((time >= timeRangeMin) & (time <= timeRangeMax))[0]

    return ind

def syncDataSeriesLagForceMoment(timeData,heightData,forceXData,forceZData,momentYData,lag):
    shiftedHeightData = [heightData[i+lag] for i in range(len(heightData)-lag)]
    shiftedForceXData = [forceXData[i+lag] for i in range(len(forceXData)-lag)]
    shiftedForceZData = [forceZData[i+lag] for i in range(len(forceZData)-lag)]
    shiftedMomentYData = [momentYData[i+lag] for i in range(len(momentYData)-lag)]
    shiftedTimeData = timeData[0:len(shiftedHeightData)]
    
    return shiftedTimeData,np.asarray(shiftedHeightData),np.asarray(shiftedForceXData),np.asarray(shiftedForceZData),np.asarray(shiftedMomentYData)

def findZeros(data):
    zeroCrossings = np.where(np.diff(np.signbit(data)))[0]
    return zeroCrossings


def findUpCrossings(heightData, timeData):
    upCrossings = []
    zcross = findZeros(heightData)

    for ind in zcross:
        # This need additional clean up
        if (np.sign(heightData[ind-1]) == -1) and (timeData[ind] - timeData[0] > 0.1):
            upCrossings.append(ind)

    return np.asarray(upCrossings)


def findHeightPeriodTrend(heightData, timeData):

    assert len(heightData) == len(timeData)

    ups = findUpCrossings(heightData, timeData)
    period = []
    height = []
    time = []

    i = 0
    while i <= (len(ups)-2):
        period.append(float(timeData[ups[i+1]]-timeData[ups[i]]))
        height.append(float(max(heightData[ups[i]:ups[i+1]])-min(heightData[ups[i]:ups[i+1]])))
        time.append(float(timeData[ups[i]]))
        i += 1

    return np.asarray(period), np.asarray(height), np.asarray(time)


if __name__ == "__main__":
    main()
